import json
from collections import namedtuple

LibraryDs = namedtuple('LibraryDs', ['name', 'version'])


class InstalledLibrariesOutputLinesPostProcessor(object):
    INSTALLED_LIBRARIES_AFTER_MODIFICATIONS_TASK = 'Final list of all installed libraries in the venv'

    def __init__(self):
        self.stop_line_processing = False
        self.next_line_contains_installed_libraries_after_modifications = False
        self.installed_libraries_after_modifications = []
        self.installed_libraries_before_modifications = []

    def post_process_line(self, output_line):
        if not self.stop_line_processing:
            if not self.next_line_contains_installed_libraries_after_modifications:
                if InstalledLibrariesOutputLinesPostProcessor.INSTALLED_LIBRARIES_AFTER_MODIFICATIONS_TASK in output_line:
                    self.next_line_contains_installed_libraries_after_modifications = True
            else:
                ip_and_command_info_parts = output_line.split(' => ')
                command_info = json.loads(ip_and_command_info_parts[1])
                installed_libraries_with_versions = command_info['stdout_lines']

                for installed_library_with_version in installed_libraries_with_versions:
                    name_and_version_parts = installed_library_with_version.split('==')
                    if name_and_version_parts[0] != 'pkg-resources':
                        self.installed_libraries_after_modifications.append(
                            LibraryDs(name=name_and_version_parts[0], version=name_and_version_parts[1]))
                self.stop_line_processing = True


class InstalledVirtualEnvironmentsOutputLinesPostProcessor(object):
    INSTALLED_VIRTUAL_ENVIRONMENTS_TASK = 'list all installed virtual environments'

    def __init__(self):
        self.stop_line_processing = False
        self.next_line_contains_installed_virtual_environments = False
        self.installed_virtual_environments = []

    def post_process_line(self, output_line):
        if not self.stop_line_processing:
            if not self.next_line_contains_installed_virtual_environments:
                if InstalledVirtualEnvironmentsOutputLinesPostProcessor.INSTALLED_VIRTUAL_ENVIRONMENTS_TASK in output_line:
                    self.next_line_contains_installed_virtual_environments = True
            else:
                ip_and_command_info_parts = output_line.split(' => ')
                command_info = json.loads(ip_and_command_info_parts[1])
                installed_virtual_envs = command_info['stdout_lines']

                for installed_virtual_env in installed_virtual_envs:
                    self.installed_virtual_environments.append(installed_virtual_env)
                self.stop_line_processing = True


class InitializationOutputLinesPostProcessor(object):
    PYTHON_VERSION_IDENTIFYING_TASK = 'Identify installed python version'

    def __init__(self):
        self.stop_line_processing = False
        self.next_line_contains_python_version = False
        self.python_version = []

    def post_process_line(self, output_line):
        if not self.stop_line_processing:
            if not self.next_line_contains_python_version:
                if InitializationOutputLinesPostProcessor.PYTHON_VERSION_IDENTIFYING_TASK in output_line:
                    self.next_line_contains_python_version = True
            else:
                ip_and_command_info_parts = output_line.split(' => ')
                command_info = json.loads(ip_and_command_info_parts[1])
                self.python_version = command_info['stdout_lines'][0].replace('Python', '')
                self.stop_line_processing = True


class NullOutputLinesPostProcessor(object):
    def post_process_line(self, output_line):
        pass



from __future__ import unicode_literals

import logging

from waldur_mastermind.invoices import registrators
from waldur_mastermind.invoices import models as invoice_models
from waldur_slurm import models as slurm_models
from . import models, utils

logger = logging.getLogger(__name__)


class AllocationRegistrator(registrators.BaseRegistrator):
    def get_sources(self, customer):
        return slurm_models.Allocation.objects.filter(service_project_link__project__customer=customer).distinct()

    def get_customer(self, source):
        return source.service_project_link.project.customer

    def _find_item(self, source, now):
        allocation = source
        result = invoice_models.GenericInvoiceItem.objects.filter(
            scope=allocation,
            invoice__customer=allocation.service_project_link.project.customer,
            invoice__state=invoice_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        ).first()
        return result

    def _create_item(self, source, invoice, start, end):
        allocation = source
        package = self.get_package(allocation)
        if package:
            return invoice_models.GenericInvoiceItem.objects.create(
                scope=allocation,
                project=source.service_project_link.project,
                unit_price=utils.get_deposit_usage(allocation, package),
                unit=invoice_models.GenericInvoiceItem.Units.QUANTITY,
                quantity=1,
                product_code=package.product_code,
                article_code=package.article_code,
                invoice=invoice,
                start=start,
                end=end,
            )

    def get_package(self, allocation):
        service_settings = allocation.service_project_link.service.settings
        try:
            return models.SlurmPackage.objects.get(service_settings=service_settings)
        except models.SlurmPackage.DoesNotExist:
            logger.debug('Skip SLURM invoice item because pricing package'
                         ' for service settings %s is not defined.', service_settings)

    def get_details(self, source):
        return {
            'cpu_usage': source.cpu_usage,
            'gpu_usage': source.gpu_usage,
            'ram_usage': source.ram_usage,
            'deposit_usage': source.deposit_usage,
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from waldur_core.quotas import models as quotas_models
from django.contrib.contenttypes.models import ContentType

from .. import models


def initialize_subnet_count_quota(apps, schema_editor):
    tenant_content_type = ContentType.objects.get_for_model(models.Tenant)
    quota_name = 'subnet_count'

    for tenant in models.Tenant.objects.exclude(quotas__name=quota_name).iterator():
        tenant.quotas.add(quotas_models.Quota(
            name=quota_name,
            limit=10,
            usage=tenant.networks.count(),
            content_type=tenant_content_type))


def initialize_network_count_quota(apps, schema_editor):
    tenant_content_type = ContentType.objects.get_for_model(models.Tenant)
    quota_name = 'network_count'

    for tenant in models.Tenant.objects.exclude(quotas__name=quota_name).iterator():
        subnet_count = models.SubNet.objects.filter(network__pk__in=tenant.networks.values('pk')).count()
        tenant.quotas.add(quotas_models.Quota(
            name=quota_name,
            limit=10,
            usage=subnet_count,
            content_type=tenant_content_type))


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0028_security_group_as_a_resource'),
    ]

    operations = [
        migrations.RunPython(initialize_network_count_quota),
        migrations.RunPython(initialize_subnet_count_quota),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-17 12:00
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers
import taggit.managers


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('structure', '0052_customer_subnets'),
    ]

    operations = [
        migrations.CreateModel(
            name='AzureService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='AzureServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cloud_service_name', models.CharField(blank=True, max_length=255)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureService')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InstanceEndpoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('local_port', models.IntegerField(validators=[django.core.validators.MaxValueValidator(65535)])),
                ('public_port', models.IntegerField(validators=[django.core.validators.MaxValueValidator(65535)])),
                ('protocol', models.CharField(blank=True, choices=[('tcp', 'tcp'), ('udp', 'udp')], max_length=3)),
                ('name', models.CharField(blank=True, choices=[('SSH', 'SSH'), ('Remote Desktop', 'Remote Desktop')], max_length=255)),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VirtualMachine',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('latitude', models.FloatField(blank=True, null=True)),
                ('longitude', models.FloatField(blank=True, null=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('cores', models.PositiveSmallIntegerField(default=0, help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(default=0, help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(default=0, help_text='Disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('image_name', models.CharField(blank=True, max_length=150)),
                ('key_name', models.CharField(blank=True, max_length=50)),
                ('key_fingerprint', models.CharField(blank=True, max_length=47)),
                ('user_data', models.TextField(blank=True, help_text='Additional data that will be added to instance on provisioning')),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('public_ips', waldur_core.core.fields.JSONField(blank=True, default=[], help_text='List of public IP addresses')),
                ('private_ips', waldur_core.core.fields.JSONField(blank=True, default=[], help_text='List of private IP addresses')),
                ('user_username', models.CharField(max_length=50)),
                ('user_password', models.CharField(max_length=50)),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='virtualmachines', to='waldur_azure.AzureServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='instanceendpoint',
            name='instance',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='endpoints', to='waldur_azure.VirtualMachine'),
        ),
        migrations.AddField(
            model_name='azureservice',
            name='projects',
            field=models.ManyToManyField(related_name='azure_services', through='waldur_azure.AzureServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='azureservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AlterUniqueTogether(
            name='azureserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='azureservice',
            unique_together=set([('customer', 'settings')]),
        ),
    ]



from __future__ import unicode_literals

from django.contrib import admin
from django.forms import ModelForm
from jsoneditor.forms import JSONEditor

from . import models


class ExpertProviderAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'customer', 'created')


class ExpertRequestModel(ModelForm):
    class Meta:
        widgets = {
            'extra': JSONEditor(),
        }


class ExpertRequestAdmin(admin.ModelAdmin):
    list_display = ('name', 'project', 'state', 'created')
    readonly_fields = ('project', 'created')
    list_filter = ('state', 'created')
    search_fields = ('name', 'uuid')
    form = ExpertRequestModel


admin.site.register(models.ExpertProvider, ExpertProviderAdmin)
admin.site.register(models.ExpertRequest, ExpertRequestAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-08 07:41
from __future__ import unicode_literals

import django.utils.timezone
import model_utils.fields
from django.db import migrations, models

import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0002_category_offering'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('key', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255)),
                ('type', models.CharField(choices=[('boolean', b'boolean'), ('string', b'string'), ('integer', b'integer'), ('choice', b'choice'), ('list', b'list')], max_length=255)),
                ('available_values', waldur_core.core.fields.JSONField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Section',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('key', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='attribute',
            name='section',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='marketplace.Section'),
        ),
        migrations.AddField(
            model_name='section',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sections', to='marketplace.Category'),
        ),
    ]



from django.utils import timezone

from waldur_mastermind.invoices import registrators

from . import utils


def add_new_allocation_to_invoice(sender, instance, created=False, **kwargs):
    if not created:
        return

    registrators.RegistrationManager.register(instance, timezone.now())


def terminate_invoice_when_allocation_deleted(sender, instance, **kwargs):
    registrators.RegistrationManager.terminate(instance, timezone.now())


def terminate_invoice_when_allocation_cancelled(sender, instance, created=False, **kwargs):
    if created:
        return

    if instance.tracker.has_changed('is_active') and not instance.is_active:
        registrators.RegistrationManager.terminate(instance, timezone.now())


def update_invoice_item_on_allocation_usage_update(sender, instance, created=False, **kwargs):
    if created:
        return

    allocation = instance
    if not allocation.usage_changed():
        return

    invoice_item = registrators.RegistrationManager.get_item(allocation)
    if not invoice_item:
        return

    package = utils.get_package(allocation)
    if package:
        invoice_item.unit_price = utils.get_deposit_usage(allocation, package)
        invoice_item.save(update_fields=['unit_price'])


def update_allocation_deposit(sender, instance, created=False, **kwargs):
    allocation = instance

    package = utils.get_package(allocation)
    if not package:
        return

    if created or allocation.usage_changed():
        if 'deposit_limit' in (kwargs.get('update_fields') or {}):
            return

        allocation.deposit_limit = utils.get_deposit_limit(allocation, package)
        if created:
            allocation.save()
        else:
            allocation.save(update_fields=['deposit_limit'])



import json

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.reverse import reverse
from rest_framework import serializers, status
from rest_framework.test import APIRequestFactory

from waldur_mastermind.packages import models as package_models
from waldur_mastermind.packages import views as package_views
from waldur_openstack.openstack import models as openstack_models


factory = APIRequestFactory()


def process_order_item(order_item, request):
    try:
        template = order_item.offering.scope
    except ObjectDoesNotExist:
        template = None

    if not isinstance(template, package_models.PackageTemplate):
        raise serializers.ValidationError('Offering has invalid scope. VPC package template is expected.')

    project = order_item.order.project

    try:
        spl = openstack_models.OpenStackServiceProjectLink.objects.get(
            project=project,
            service__settings=template.service_settings,
            service__customer=project.customer,
        )
    except openstack_models.OpenStackServiceProjectLink.DoesNotExist:
        raise serializers.ValidationError('Project does not have access to the OpenStack service.')

    project_url = reverse('project-detail', kwargs={'uuid': project.uuid})
    spl_url = reverse('openstack-spl-detail', kwargs={'pk': spl.pk}, request=request)
    template_url = reverse('package-template-detail', kwargs={'uuid': template.uuid})

    post_data = dict(
        name=order_item.attributes['name'],
        description=order_item.attributes.get('description'),
        user_username=order_item.attributes.get('user_username'),
        service_project_link=spl_url,
        project=project_url,
        template=template_url,
    )
    request = factory.post('/', data=json.dumps(post_data), content_type='application/json',
                           HTTP_AUTHORIZATION='Token %s' % request.user.auth_token.key)

    view = package_views.OpenStackPackageViewSet.as_view({'post': 'create'})
    response = view(request)
    if response.status_code != status.HTTP_201_CREATED:
        raise serializers.ValidationError(response.data)

    package_uuid = response.data['uuid']
    package = package_models.OpenStackPackage.objects.get(uuid=package_uuid)
    order_item.scope = package
    order_item.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-19 13:04
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0052_customer_subnets'),
        ('waldur_jira', '0003_project_template'),
    ]

    operations = [
        migrations.CreateModel(
            name='IssueType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('icon_url', models.URLField(blank=True, verbose_name='icon url')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Issue type',
                'verbose_name_plural': 'Issue types',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.RemoveField(
            model_name='project',
            name='default_issue_type',
        ),
        migrations.AlterField(
            model_name='issue',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_jira.IssueType'),
        ),
        migrations.AddField(
            model_name='issuetype',
            name='projects',
            field=models.ManyToManyField(related_name='issue_types', to='waldur_jira.Project'),
        ),
        migrations.AddField(
            model_name='issuetype',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
        ),
    ]



from celery import shared_task

from . import cost_tracking, openstack, slurm, utils


@shared_task(name='analytics.push_points')
def push_points():
    client = utils.get_influxdb_client()
    if not client:
        return
    points = []
    points.extend(openstack.get_tenants())
    points.extend(openstack.get_instances())
    points.extend(cost_tracking.get_total_cost())
    points.extend(slurm.get_usage())
    utils.write_points(client, points)



from __future__ import unicode_literals

from django.db import transaction
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_mastermind.invoices import registrators as invoices_registrators

from . import models, tasks, quotas
from .log import event_logger


def update_project_quota_when_request_is_saved(sender, instance, created=False, **kwargs):
    if created or instance.tracker.has_changed('state'):
        transaction.on_commit(lambda: quotas.update_project_quota(instance.project))


def update_project_quota_when_request_is_deleted(sender, instance, **kwargs):
    transaction.on_commit(lambda: quotas.update_project_quota(instance.project))


def update_customer_quota_when_request_is_saved(sender, instance, created=False, **kwargs):
    if created or instance.tracker.has_changed('state'):
        customer = quotas.get_request_customer(instance)
        transaction.on_commit(lambda: quotas.update_customer_quota(customer))


def update_customer_quota_when_request_is_deleted(sender, instance, **kwargs):
    customer = quotas.get_request_customer(instance)
    transaction.on_commit(lambda: quotas.update_customer_quota(customer))


def update_customer_quota_when_contract_is_created(sender, instance, created=False, **kwargs):
    if created:
        customer = quotas.get_contract_customer(instance)
        transaction.on_commit(lambda: quotas.update_customer_quota(customer))


def update_customer_quota_when_contract_is_deleted(sender, instance, **kwargs):
    customer = quotas.get_contract_customer(instance)
    transaction.on_commit(lambda: quotas.update_customer_quota(customer))


def log_expert_request_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    event_logger.waldur_expert_request.info(
        'User {user_username} with full name {user_full_name} has created '
        'request for experts under {customer_name} / {project_name}.',
        event_type='expert_request_created',
        event_context={
            'expert_request': instance,
        })


def log_expert_request_state_changed(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    if instance.state == models.ExpertRequest.States.ACTIVE:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been activated.',
            event_type='expert_request_activated',
            event_context={
                'expert_request': instance,
            })
    elif instance.state == models.ExpertRequest.States.CANCELLED:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been cancelled.',
            event_type='expert_request_cancelled',
            event_context={
                'expert_request': instance,
            })
    elif instance.state == models.ExpertRequest.States.COMPLETED:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been completed.',
            event_type='expert_request_completed',
            event_context={
                'expert_request': instance,
            })


def log_expert_bid_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    event_logger.waldur_expert_bid.info(
        'User {user_username} with full name {user_full_name} has created '
        'bid for request {request_name} under {customer_name} / {project_name}.',
        event_type='expert_bid_created',
        event_context={
            'expert_bid': instance,
        })


def add_completed_expert_request_to_invoice(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state != models.ExpertRequest.States.COMPLETED or state == instance.tracker.previous('state'):
        return

    if not instance.issue or not hasattr(instance, 'contract'):
        return

    invoices_registrators.RegistrationManager.register(instance, timezone.now())


def terminate_invoice_when_expert_request_deleted(sender, instance, **kwargs):
    invoices_registrators.RegistrationManager.terminate(instance, timezone.now())


def set_project_name_on_expert_request_creation(sender, instance, created=False, **kwargs):
    if created:
        request = instance
        request.project_name = request.project.name
        request.project_uuid = request.project.uuid.hex
        request.customer = request.project.customer
        request.save(update_fields=('project_name', 'project_uuid', 'customer'))


def update_expert_request_on_project_name_update(sender, instance, **kwargs):
    project = instance
    if project.tracker.has_changed('name'):
        models.ExpertRequest.objects.filter(project=project).update(project_name=project.name)


def set_team_name_on_expert_contract_creation(sender, instance, created=False, **kwargs):
    if created:
        contract = instance
        contract.team_name = contract.team.name
        contract.team_uuid = contract.team.uuid.hex
        contract.team_customer = contract.team.customer
        contract.save(update_fields=('team_name', 'team_uuid', 'team_customer'))


def update_expert_contract_on_project_name_update(sender, instance, **kwargs):
    project = instance
    if project.tracker.has_changed('name'):
        models.ExpertContract.objects.filter(team=project).update(team_name=project.name)


def notify_expert_providers_about_new_request(sender, instance, created=False, **kwargs):
    if created:
        transaction.on_commit(lambda:
                              tasks.send_new_request.delay(instance.uuid.hex))


def notify_customer_owners_about_new_bid(sender, instance, created=False, **kwargs):
    if created:
        transaction.on_commit(lambda:
                              tasks.send_new_bid.delay(instance.uuid.hex))


def notify_customer_owners_about_new_contract(sender, instance, created=False, **kwargs):
    if created:
        transaction.on_commit(lambda:
                              tasks.send_new_contract.delay(instance.request.uuid.hex))


def send_expert_comment_added_notification(sender, instance, created=False, **kwargs):
    # Send Expert notifications
    comment = instance

    if not created or not comment.is_public:
        return

    serialized_comment = core_utils.serialize_instance(comment)
    transaction.on_commit(lambda:
                          tasks.send_expert_comment_added_notification.delay(serialized_comment))



from celery import shared_task

from . import utils
from .backend import FreeIPABackend


def schedule_sync():
    """
    This function calls task only if it is not already running.
    The goal is to avoid race conditions during concurrent task execution.
    """
    if utils.is_syncing():
        return

    utils.renew_task_status()
    _sync_groups.apply_async(countdown=10)


@shared_task(name='waldur_freeipa.sync_groups')
def sync_groups():
    """
    This task is used by Celery beat in order to periodically
    schedule FreeIPA group synchronization.
    """
    schedule_sync()


@shared_task()
def _sync_groups():
    """
    This task actually calls backend. It is called asynchronously
    either by signal handler or Celery beat schedule.
    """
    FreeIPABackend().synchronize_groups()


def schedule_sync_names():
    _sync_names.apply_async(countdown=10)


@shared_task()
def _sync_names():
    FreeIPABackend().synchronize_names()


def schedule_sync_gecos():
    _sync_gecos.apply_async(countdown=10)


@shared_task()
def _sync_gecos():
    FreeIPABackend().synchronize_gecos()



import calendar
from collections import OrderedDict
import datetime
import importlib
from itertools import chain
from operator import itemgetter
import os
import re
import time

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.mail import send_mail
from django.core.management import call_command
from django.http import QueryDict
from django.template.loader import render_to_string
from django.urls import resolve
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.encoding import force_text


def flatten(*xs):
    return tuple(chain.from_iterable(xs))


def sort_dict(unsorted_dict):
    """
    Return a OrderedDict ordered by key names from the :unsorted_dict:
    """
    sorted_dict = OrderedDict()
    # sort items before inserting them into a dict
    for key, value in sorted(unsorted_dict.items(), key=itemgetter(0)):
        sorted_dict[key] = value
    return sorted_dict


def format_time_and_value_to_segment_list(time_and_value_list, segments_count, start_timestamp,
                                          end_timestamp, average=False):
    """
    Format time_and_value_list to time segments

    Parameters
    ^^^^^^^^^^
    time_and_value_list: list of tuples
        Have to be sorted by time
        Example: [(time, value), (time, value) ...]
    segments_count: integer
        How many segments will be in result
    Returns
    ^^^^^^^
    List of dictionaries
        Example:
        [{'from': time1, 'to': time2, 'value': sum_of_values_from_time1_to_time2}, ...]
    """
    segment_list = []
    time_step = (end_timestamp - start_timestamp) / segments_count
    for i in range(segments_count):
        segment_start_timestamp = start_timestamp + time_step * i
        segment_end_timestamp = segment_start_timestamp + time_step
        value_list = [
            value for time, value in time_and_value_list
            if time >= segment_start_timestamp and time < segment_end_timestamp]
        segment_value = sum(value_list)
        if average and len(value_list) != 0:
            segment_value /= len(value_list)

        segment_list.append({
            'from': segment_start_timestamp,
            'to': segment_end_timestamp,
            'value': segment_value,
        })
    return segment_list


def datetime_to_timestamp(datetime):
    return int(time.mktime(datetime.timetuple()))


def timestamp_to_datetime(timestamp, replace_tz=True):
    dt = datetime.datetime.fromtimestamp(int(timestamp))
    if replace_tz:
        dt = dt.replace(tzinfo=timezone.get_current_timezone())
    return dt


def timeshift(**kwargs):
    return timezone.now().replace(microsecond=0) + datetime.timedelta(**kwargs)


def hours_in_month(month=None, year=None):
    now = datetime.datetime.now()
    if not month:
        month = now.month
    if not year:
        year = now.year

    days_in_month = calendar.monthrange(year, month)[1]
    return 24 * days_in_month


def month_start(date):
    return timezone.make_aware(datetime.datetime(day=1, month=date.month, year=date.year))


def month_end(date):
    days_in_month = calendar.monthrange(date.year, date.month)[1]
    last_day_of_month = datetime.date(month=date.month, year=date.year, day=days_in_month)
    last_second_of_month = datetime.datetime.combine(last_day_of_month, datetime.time.max)
    return timezone.make_aware(last_second_of_month, timezone.get_current_timezone())


def pwgen(pw_len=16):
    """ Generate a random password with the given length.
        Allowed chars does not have "I" or "O" or letters and
        digits that look similar -- just to avoid confusion.
    """
    return get_random_string(pw_len, 'abcdefghjkmnpqrstuvwxyz'
                                     'ABCDEFGHJKLMNPQRSTUVWXYZ'
                                     '23456789')


def serialize_instance(instance):
    """ Serialize Django model instance """
    model_name = force_text(instance._meta)
    return '{}:{}'.format(model_name, instance.pk)


def deserialize_instance(serialized_instance):
    """ Deserialize Django model instance """
    model_name, pk = serialized_instance.split(':')
    model = apps.get_model(model_name)
    return model._default_manager.get(pk=pk)


def serialize_class(cls):
    """ Serialize Python class """
    return '{}:{}'.format(cls.__module__, cls.__name__)


def deserialize_class(serilalized_cls):
    """ Deserialize Python class """
    module_name, cls_name = serilalized_cls.split(':')
    module = importlib.import_module(module_name)
    return getattr(module, cls_name)


def clear_url(url):
    """ Remove domain and protocol from url """
    if url.startswith('http'):
        return '/' + url.split('/', 3)[-1]
    return url


def get_model_from_resolve_match(match):
    queryset = match.func.cls.queryset
    if queryset is not None:
        return queryset.model
    else:
        return match.func.cls.model


def instance_from_url(url, user=None):
    """ Restore instance from URL """
    # XXX: This circular dependency will be removed then filter_queryset_for_user
    # will be moved to model manager method
    from waldur_core.structure.managers import filter_queryset_for_user

    url = clear_url(url)
    match = resolve(url)
    model = get_model_from_resolve_match(match)
    queryset = model.objects.all()
    if user is not None:
        queryset = filter_queryset_for_user(model.objects.all(), user)
    return queryset.get(**match.kwargs)


def get_detail_view_name(model):
    if model is NotImplemented:
        raise AttributeError('Cannot get detail view name for not implemented model')

    if hasattr(model, 'get_url_name') and callable(model.get_url_name):
        return '%s-detail' % model.get_url_name()

    return '%s-detail' % model.__name__.lower()


def get_list_view_name(model):
    if model is NotImplemented:
        raise AttributeError('Cannot get list view name for not implemented model')

    if hasattr(model, 'get_url_name') and callable(model.get_url_name):
        return '%s-list' % model.get_url_name()

    return '%s-list' % model.__name__.lower()


def get_fake_context():
    user = get_user_model()()
    request = type('R', (object,), {'method': 'GET', 'user': user, 'query_params': QueryDict()})
    return {'request': request, 'user': user}


def camel_case_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def silent_call(name, *args, **options):
    call_command(name, stdout=open(os.devnull, 'w'), *args, **options)


def broadcast_mail(app, event_type, context, recipient_list):
    """
    Shorthand to format email message from template file and sent it to all recipients.

    It is assumed that there are there are 3 templates available for event type in application.
    For example, if app is 'experts' and event_type is 'new_request', then there should be 3 files:

    1) experts/new_request_subject.txt is template for email subject
    2) experts/new_request_message.txt is template for email body as text
    3) experts/new_request_message.html is template for email body as HTML

    By default, built-in Django send_mail is used, all members
    of the recipient list will see the other recipients in the 'To' field.
    Contrary to this, we're using explicit loop in order to ensure that
    recipients would NOT see the other recipients.

    :param app: prefix for template filename.
    :param event_type: postfix for template filename.
    :param context: dictionary passed to the template for rendering.
    :param recipient_list: list of strings, each an email address.
    """

    subject_template = '%s/%s_subject.txt' % (app, event_type)
    text_template = '%s/%s_message.txt' % (app, event_type)
    html_template = '%s/%s_message.html' % (app, event_type)

    subject = render_to_string(subject_template, context).strip()
    text_message = render_to_string(text_template, context)
    html_message = render_to_string(html_template, context)

    for recipient in recipient_list:
        send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [recipient], html_message=html_message)



import itertools
from celery import shared_task

from waldur_core.core import utils as core_utils
from waldur_core.structure import models as structure_models
from . import models


def get_user_allocations(user):
    project_permissions = structure_models.ProjectPermission.objects.filter(user=user, is_active=True)
    projects = project_permissions.values_list('project_id', flat=True)
    project_allocations = models.Allocation.objects.filter(is_active=True, service_project_link__project__in=projects)

    customer_permissions = structure_models.CustomerPermission.objects.filter(user=user, is_active=True)
    customers = customer_permissions.values_list('customer_id', flat=True)
    customer_allocations = models.Allocation.objects.filter(is_active=True, service_project_link__project__customer__in=customers)

    return itertools.chain(project_allocations, customer_allocations)


def get_structure_allocations(structure):
    if isinstance(structure, structure_models.Project):
        return list(models.Allocation.objects.filter(is_active=True, service_project_link__project=structure))
    elif isinstance(structure, structure_models.Customer):
        return list(models.Allocation.objects.filter(is_active=True, service_project_link__project__customer=structure))
    else:
        return []


@shared_task(name='waldur_slurm.add_user')
def add_user(serialized_profile):
    profile = core_utils.deserialize_instance(serialized_profile)
    for allocation in get_user_allocations(profile.user):
        allocation.get_backend().add_user(allocation, profile.username)


@shared_task(name='waldur_slurm.delete_user')
def delete_user(serialized_profile):
    profile = core_utils.deserialize_instance(serialized_profile)
    for allocation in get_user_allocations(profile.user):
        allocation.get_backend().delete_user(allocation, profile.username)


@shared_task(name='waldur_slurm.process_role_granted')
def process_role_granted(serialized_profile, serialized_structure):
    profile = core_utils.deserialize_instance(serialized_profile)
    structure = core_utils.deserialize_instance(serialized_structure)

    allocations = get_structure_allocations(structure)

    for allocation in allocations:
        allocation.get_backend().add_user(allocation, profile.username)


@shared_task(name='waldur_slurm.process_role_revoked')
def process_role_revoked(serialized_profile, serialized_structure):
    profile = core_utils.deserialize_instance(serialized_profile)
    structure = core_utils.deserialize_instance(serialized_structure)

    allocations = get_structure_allocations(structure)

    for allocation in allocations:
        allocation.get_backend().delete_user(allocation, profile.username)



from waldur_core.core.models import User
from waldur_core.logging.loggers import EventLogger, event_logger


class Saml2AuthEventLogger(EventLogger):
    user = User

    class Meta:
        event_types = ('auth_logged_in_with_saml2', 'auth_logged_out_with_saml2')
        event_groups = {'users': event_types}


event_logger.register('saml2_auth', Saml2AuthEventLogger)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.logging.loggers
import model_utils.fields
import waldur_core.core.fields
import waldur_core.structure.models
import waldur_core.core.models
import django.db.models.deletion
import django.utils.timezone
import taggit.managers
import django_fsm
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('openstack_tenant', '0003_volume_and_snapshot'),
    ]

    operations = [
        migrations.CreateModel(
            name='Instance',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('latitude', models.FloatField(null=True, blank=True)),
                ('longitude', models.FloatField(null=True, blank=True)),
                ('runtime_state', models.CharField(max_length=150, verbose_name='runtime state', blank=True)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')])),
                ('cores', models.PositiveSmallIntegerField(default=0, help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(default=0, help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(default=0, help_text='Disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('external_ips', models.GenericIPAddressField(null=True, protocol='IPv4', blank=True)),
                ('internal_ips', models.GenericIPAddressField(null=True, protocol='IPv4', blank=True)),
                ('image_name', models.CharField(max_length=150, blank=True)),
                ('key_name', models.CharField(max_length=50, blank=True)),
                ('key_fingerprint', models.CharField(max_length=47, blank=True)),
                ('user_data', models.TextField(help_text='Additional data that will be added to instance on provisioning', blank=True)),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('flavor_name', models.CharField(max_length=255, blank=True)),
                ('flavor_disk', models.PositiveIntegerField(default=0, help_text=b'Flavor disk size in MiB')),
                ('security_groups', models.ManyToManyField(related_name='instances', to='openstack_tenant.SecurityGroup')),
                ('service_project_link', models.ForeignKey(related_name='instances', on_delete=django.db.models.deletion.PROTECT, to='openstack_tenant.OpenStackTenantServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='volume',
            name='instance',
            field=models.ForeignKey(related_name='volumes', blank=True, to='openstack_tenant.Instance', null=True),
        ),
    ]



from __future__ import unicode_literals

from waldur_mastermind.experts import views


def register_in(router):
    router.register(r'expert-providers', views.ExpertProviderViewSet, base_name='expert-provider')
    router.register(r'expert-requests', views.ExpertRequestViewSet, base_name='expert-request')
    router.register(r'expert-bids', views.ExpertBidViewSet, base_name='expert-bid')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-14 15:37
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0021_remove_pending_offering_item'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='paymentdetails',
            name='customer',
        ),
        migrations.DeleteModel(
            name='PaymentDetails',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-23 12:48
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers
import waldur_core.structure.models
import taggit.managers


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('structure', '0052_customer_subnets'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(max_length=255)),
                ('file', models.FileField(upload_to=b'jira_attachments')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(max_length=255)),
                ('message', models.TextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(max_length=255)),
                ('type', models.CharField(max_length=255)),
                ('summary', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True)),
                ('resolution', models.CharField(blank=True, max_length=255)),
                ('priority', models.SmallIntegerField(choices=[(0, b'n/a'), (1, b'Minor'), (2, b'Major'), (3, b'Critical')], default=0)),
                ('impact', models.SmallIntegerField(choices=[(0, b'n/a'), (1, b'Small - Partial loss of service, one person affected'), (2, b'Medium - One department or service is affected'), (3, b'Large - Whole organization or all services are affected')], default=0)),
                ('status', models.CharField(max_length=255)),
                ('updated', models.DateTimeField(auto_now_add=True)),
                ('updated_username', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='JiraService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='JiraServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('impact_field', models.CharField(blank=True, max_length=64)),
                ('reporter_field', models.CharField(blank=True, max_length=64)),
                ('default_issue_type', models.CharField(blank=True, max_length=64)),
                ('available_for_all', models.BooleanField(default=False, help_text=b'Allow access to any user')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='projects', to='waldur_jira.JiraServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='jiraserviceprojectlink',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project'),
        ),
        migrations.AddField(
            model_name='jiraserviceprojectlink',
            name='service',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_jira.JiraService'),
        ),
        migrations.AddField(
            model_name='jiraservice',
            name='projects',
            field=models.ManyToManyField(related_name='jira_services', through='waldur_jira.JiraServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='jiraservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='issue',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='issues', to='waldur_jira.Project'),
        ),
        migrations.AddField(
            model_name='issue',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='comment',
            name='issue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments', to='waldur_jira.Issue'),
        ),
        migrations.AddField(
            model_name='comment',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='attachment',
            name='issue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='waldur_jira.Issue'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='jiraserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='jiraservice',
            unique_together=set([('customer', 'settings')]),
        ),
    ]



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets, decorators, response, status

from waldur_core.core import views as core_views
from waldur_core.structure import filters as structure_filters, permissions as structure_permissions

from . import models, serializers, filters, optimizers


class DeploymentPlanViewSet(core_views.ActionsViewSet):
    queryset = models.DeploymentPlan.objects.all()
    serializer_class = serializers.DeploymentPlanSerializer
    lookup_field = 'uuid'
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.DeploymentPlanFilter
    unsafe_methods_permissions = [structure_permissions.is_administrator]

    def retrieve(self, request, *args, **kwargs):
        """
        Example rendering of deployment plan and configuration.

        .. code-block:: javascript

            GET /api/deployment-plans/c218cbb2f56c4d52a82638ca9fffd85a/
            Accept: application/json
            Content-Type: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "url": "http://example.com/api/deployment-plans/b12bb98a661749ffb02c8a8439299288/",
                "uuid": "b12bb98a661749ffb02c8a8439299288",
                "name": "Webapp for Monster Inc",
                "customer": "http://example.com/api/customers/790b3c131e894581b3dcf66796d9fa30/",
                "items": [
                    {
                        "preset": {
                            "url": "http://example.com/api/deployment-presets/628cd853ba2a4ce7af5d4fff510b5bd2/",
                            "uuid": "628cd853ba2a4ce7af5d4fff510b5bd2",
                            "name": "MySQL",
                            "category": "Databases",
                            "variant": "Large",
                            "ram": 2048,
                            "cores": 2,
                            "storage": 1024000
                        },
                        "quantity": 1
                    }
                ]
            }
        """
        return super(DeploymentPlanViewSet, self).retrieve(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        Example request for creating deployment plan.

        .. code-block:: javascript

            POST /api/deployment-plans/
            Accept: application/json
            Content-Type: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "WebApps",
                "customer": "http://example.com/api/customers/2f8b4e0f101545508d52c7655d6386c8/",
                "items": [
                    {
                        "preset": "http://example.com/api/deployment-presets/2debb6d109954afaa03910ba1c6791a6/",
                        "quantity": 1
                    }
                ]
            }
        """
        return super(DeploymentPlanViewSet, self).create(request, *args, **kwargs)

    create_serializer_class = serializers.DeploymentPlanCreateSerializer

    def update(self, request, *args, **kwargs):
        """
        Run **PUT** request against */api/deployment-plans/<uuid>/* to update deployment plan.
        Only name and list of items can be updated.
        List of items should have the same format as POST request.
        Only customer owner and staff can update deployment plan.
        """
        return super(DeploymentPlanViewSet, self).update(request, *args, **kwargs)

    update_serializer_class = partial_update_serializer_class = serializers.DeploymentPlanCreateSerializer

    @decorators.detail_route(methods=['GET'])
    def evaluate(self, request, *args, **kwargs):
        strategy = optimizers.SingleServiceStrategy(self.get_object())
        optimized_services = strategy.get_optimized()
        serializer = self.get_serializer(optimized_services, many=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)

    evaluate_serializer_class = serializers.OptimizedServiceSummarySerializer


class PresetViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Preset.objects.all()
    serializer_class = serializers.PresetSerializer
    lookup_field = 'uuid'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
from django.conf import settings
import model_utils.fields
import waldur_core.core.fields
import waldur_core.structure.models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('structure', '0037_remove_customer_billing_backend_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('key', models.CharField(max_length=255)),
                ('type', models.CharField(default='informational', max_length=30, choices=[('informational', 'Informational'), ('service_request', 'Service request'), ('change_request', 'Change request'), ('incident', 'Incident')])),
                ('summary', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True)),
                ('status', models.CharField(max_length=255)),
                ('resolution', models.CharField(max_length=255, blank=True)),
                ('resource_object_id', models.PositiveIntegerField(null=True)),
                ('assignee', models.ForeignKey(related_name='assigned_issues', blank=True, to=settings.AUTH_USER_MODEL, null=True)),
                ('creator', models.ForeignKey(related_name='created_issues', to=settings.AUTH_USER_MODEL)),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='issues', blank=True, to='structure.Customer', null=True)),
                ('project', models.ForeignKey(related_name='issues', blank=True, to='structure.Project', null=True)),
                ('reporter', models.ForeignKey(related_name='reported_issues', to=settings.AUTH_USER_MODEL)),
                ('resource_content_type', models.ForeignKey(to='contenttypes.ContentType', null=True)),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(models.Model, waldur_core.structure.models.StructureLoggableMixin),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-12 09:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0018_offering_terminated_at'),
    ]

    operations = [
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.TextField()),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('native_name', models.CharField(blank=True, max_length=150)),
                ('native_description', models.TextField(blank=True)),
                ('issue_type', models.CharField(choices=[('INFORMATIONAL', 'Informational'), ('SERVICE_REQUEST', 'Service request'), ('CHANGE_REQUEST', 'Change request'), ('INCIDENT', 'Incident')], default='INFORMATIONAL', max_length=30)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TemplateAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('file', models.FileField(upload_to='support_template_attachments')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='support.Template')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import collections
import logging
import os
import re
import ssl
import tempfile
import time

from django.core.files.uploadedfile import File, InMemoryUploadedFile
from django.db import IntegrityError
from django.utils import six
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.compute.base import NodeAuthPassword
from libcloud.compute.drivers import azure
from libcloud.compute.types import NodeState

from waldur_core.structure import ServiceBackend, ServiceBackendError, ServiceBackendNotImplemented, \
    log_backend_action

from . import models
from .driver import AzureNodeDriver, AZURE_COMPUTE_INSTANCE_TYPES


logger = logging.getLogger(__name__)

# libcloud doesn't match Visual Studio images properly
azure.WINDOWS_SERVER_REGEX = re.compile(
    azure.WINDOWS_SERVER_REGEX.pattern + '|VS-201[35]'
)

# there's a hope libcloud will implement this method in further releases
AzureNodeDriver.ex_list_storage_services = lambda self: \
    self._perform_get(self._get_path('services', 'storageservices'), StorageServices)


class SizeQueryset(object):
    def __init__(self):
        self.items = []
        for key, val in AZURE_COMPUTE_INSTANCE_TYPES.items():
            self.items.append(SizeQueryset.Size(uuid=val['id'],
                                                pk=val['id'],
                                                name='{}: {}'.format(key, val['name']),
                                                cores=isinstance(val['cores'], int) and val['cores'] or 1,
                                                ram=val['ram'],
                                                disk=ServiceBackend.gb2mb(val['disk']),
                                                price=float(val['price'])))

        self.items = list(sorted(self.items, key=lambda s: s.price))

    def __len__(self):
        return len(self.items)

    def __getitem__(self, key):
        return self.items[key]

    def all(self):
        return self.items

    def get(self, uuid):
        for item in self.items:
            if item.uuid == uuid:
                return item
        raise ValueError

    class Size(collections.namedtuple('Size', ('uuid', 'pk', 'name', 'cores', 'ram', 'disk', 'price'))):
        def __str__(self):
            return self.name


class StorageServiceProperties(azure.WindowsAzureData):

    def __init__(self):
        self.description = ''
        self.location = ''
        self.affinity_group = ''
        self.label = azure._Base64String()
        self.status = ''
        self.createion_time = ''


class StorageService(azure.WindowsAzureData):
    _repr_attributes = [
        'service_name',
        'url'
    ]

    def __init__(self):
        self.url = ''
        self.service_name = ''
        self.storage_service_properties = StorageServiceProperties()


class StorageServices(azure.WindowsAzureDataTypedList, azure.ReprMixin):
    list_type = StorageService

    _repr_attributes = [
        'items'
    ]


class AzureBackendError(ServiceBackendError):
    pass


class AzureBaseBackend(ServiceBackend):

    State = NodeState

    def __init__(self, settings, cloud_service_name=None):
        self.deployment = 'production'
        self.location = 'Central US'
        if settings.options and 'location' in settings.options:
            self.location = settings.options['location']

        self.settings = settings
        self.cloud_service_name = cloud_service_name

    def __del__(self):
        try:
            os.remove(self._temp_file)
        except (AttributeError, OSError):
            pass

    # Lazy init
    @property
    def manager(self):
        if not hasattr(self, '_manager'):
            key_file = None
            cert_file = self.settings.certificate.file if self.settings.certificate else None
            if isinstance(cert_file, InMemoryUploadedFile):
                cert_file.seek(0)
                temp_file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
                temp_file.writelines(cert_file.readlines())
                temp_file.close()
                cert_file.seek(0)
                key_file = self._temp_file = temp_file.name
            elif isinstance(cert_file, File):
                key_file = cert_file.name

            try:
                self._manager = AzureNodeDriver(
                    subscription_id=self.settings.username, key_file=key_file)
            except InvalidCredsError as e:
                logger.exception("Wrong credentials for service settings %s", self.settings.uuid)
                six.reraise(AzureBackendError, e)
        return self._manager

    def sync(self):
        self.pull_service_properties()

    def sync_link(self, service_project_link, is_initial=False):
        self.push_link(service_project_link)

    def remove_link(self, service_project_link):
        # TODO: this should remove storage and cloud service
        raise ServiceBackendNotImplemented


class AzureBackend(AzureBaseBackend):
    """ Waldur interface to Azure API.
        http://libcloud.readthedocs.org/en/latest/compute/drivers/azure.html
    """

    def ping(self, raise_exception=False):
        try:
            self.manager.list_locations()
        except (AzureBackendError, LibcloudError, ssl.SSLError) as e:
            if raise_exception:
                six.reraise(AzureBackendError, e)
            return False
        else:
            return True

    def ping_resource(self, instance):
        try:
            self.get_vm(instance.backend_id)
        except AzureBackendError:
            return False
        else:
            return True

    def pull_service_properties(self):
        self.pull_images()

    def pull_images(self):
        options = self.settings.options or {}
        regex = None
        if 'images_regex' in options:
            try:
                regex = re.compile(options['images_regex'])
            except re.error:
                logger.warning(
                    'Invalid images regexp supplied for service settings %s: %s',
                    self.settings.uuid, options['images_regex'])

        images = {}
        for image in self.manager.list_images():
            images.setdefault(image.name, [])
            images[image.name].append(image)

        cur_images = {i.backend_id: i for i in models.Image.objects.all()}
        for backend_images in images.values():
            backend_image = sorted(backend_images)[-1]  # get last image with same name (perhaps newest one)
            if regex and not regex.match(backend_image.name):
                continue
            cur_images.pop(backend_image.id, None)
            try:
                models.Image.objects.update_or_create(
                    backend_id=backend_image.id,
                    defaults={
                        'name': backend_image.name,
                    })
            except IntegrityError:
                logger.warning(
                    'Could not create Azure image with id %s due to concurrent update',
                    backend_image.id)

        map(lambda i: i.delete(), cur_images.values())

    def push_link(self, service_project_link):
        # define cloud service name
        options = service_project_link.service.settings.options
        if options and 'cloud_service_name' in options:
            cloud_service_name = options['cloud_service_name']
            service_project_link.cloud_service_name = cloud_service_name
            service_project_link.save(update_fields=['cloud_service_name'])
        else:
            cloud_service_name = 'nc-%x' % service_project_link.project.uuid.node

        # create cloud
        services = [s.service_name for s in self.manager.ex_list_cloud_services()]
        if cloud_service_name not in services:
            logger.debug('About to create new azure cloud service for SPL %s', service_project_link.pk)
            self.manager.ex_create_cloud_service(cloud_service_name, self.location)
            service_project_link.cloud_service_name = cloud_service_name
            service_project_link.save(update_fields=['cloud_service_name'])
            logger.info('Successfully created new azure cloud for SPL %s', service_project_link.pk)
        else:
            logger.debug('Skipped azure cloud service creation for SPL %s - such cloud already exists', service_project_link.pk)

        # create storage
        storage_name = self.get_storage_name(cloud_service_name)
        storages = [s.service_name for s in self.manager.ex_list_storage_services()]

        if storage_name not in storages:
            logger.debug('About to create new azure storage for SPL %s', service_project_link.pk)
            self.manager.ex_create_storage_service(storage_name, self.location)

            # XXX: missed libcloud feature
            #      it will block celery worker for a while (5 min max)
            #      but it's easiest workaround for azure and general syncing workflow
            for _ in range(100):
                storage = self.get_storage(storage_name)
                if storage.storage_service_properties.status == 'Created':  # ResolvingDns otherwise
                    break
                time.sleep(30)
            logger.info('Successfully created new azure storage for SPL %s', service_project_link.pk)
        else:
            logger.debug(
                'Skipped azure storage creation for SPL %s - such cloud already exists', service_project_link.pk)

    @log_backend_action()
    def reboot_vm(self, vm):
        self.manager.reboot_node(
            self.get_vm(vm.backend_id),
            ex_cloud_service_name=self.cloud_service_name,
            ex_deployment_slot=self.deployment)

    @log_backend_action()
    def stop_vm(self, vm):
        deployment_name = self.manager._get_deployment(
            service_name=self.cloud_service_name,
            deployment_slot=self.deployment
        ).name

        try:
            response = self.manager._perform_post(
                self.manager._get_deployment_path_using_name(
                    self.cloud_service_name, deployment_name
                ) + '/roleinstances/' + azure._str(vm.backend_id) + '/Operations',
                azure.AzureXmlSerializer.shutdown_role_operation_to_xml()
            )

            self.manager.raise_for_response(response, 202)
            self.manager._ex_complete_async_azure_operation(response)
        except Exception as e:
            six.reraise(AzureBackendError, e)

    @log_backend_action()
    def start_vm(self, vm):
        deployment_name = self.manager._get_deployment(
            service_name=self.cloud_service_name,
            deployment_slot=self.deployment
        ).name

        try:
            response = self.manager._perform_post(
                self.manager._get_deployment_path_using_name(
                    self.cloud_service_name, deployment_name
                ) + '/roleinstances/' + azure._str(vm.backend_id) + '/Operations',
                azure.AzureXmlSerializer.start_role_operation_to_xml()
            )

            self.manager.raise_for_response(response, 202)
            self.manager._ex_complete_async_azure_operation(response)
        except Exception as e:
            six.reraise(AzureBackendError, e)

    @log_backend_action()
    def destroy_vm(self, vm):
        self.manager.destroy_node(
            self.get_vm(vm.backend_id),
            ex_cloud_service_name=self.cloud_service_name,
            ex_deployment_slot=self.deployment)

    @log_backend_action('check if virtual machine deleted')
    def is_vm_deleted(self, vm):
        try:
            self.get_vm(vm.backend_id)
        except AzureBackendError:
            return True
        else:
            return False

    @log_backend_action()
    def provision_vm(self, vm, backend_image_id=None, backend_size_id=None):
        try:
            backend_vm = self.manager.create_node(
                name=vm.name,
                size=self.get_size(backend_size_id),
                image=self.get_image(backend_image_id),
                ex_cloud_service_name=self.cloud_service_name,
                ex_storage_service_name=self.get_storage_name(),
                ex_deployment_slot=self.deployment,
                ex_custom_data=vm.user_data,
                ex_admin_user_id=vm.user_username,
                auth=NodeAuthPassword(vm.user_password))
        except LibcloudError as e:
            logger.exception('Failed to provision virtual machine %s', vm.name)
            six.reraise(AzureBackendError, e)

        vm.backend_id = backend_vm.id
        vm.runtime_state = backend_vm.state
        vm.save(update_fields=['backend_id', 'runtime_state'])

    def pull_virtual_machine_runtime_state(self, vm):
        backend_vm = self.get_vm(vm.backend_id)
        vm.runtime_state = backend_vm.state
        vm.save(update_fields=['runtime_state'])

    def pull_vm_info(self, vm):
        """
        VM network info os available only after instance has been initiated and started.
        :param vm: waldur virtual machine instance to update IPs
        """
        backend_vm = self.get_vm(vm.backend_id)
        endpoints = []
        for endpoint in backend_vm.extra.get('instance_endpoints'):
            endpoints.append(models.InstanceEndpoint(
                name=endpoint.name,
                local_port=int(endpoint.local_port),
                public_port=int(endpoint.public_port),
                protocol=endpoint.protocol,
                instance=vm,
            ))
        models.InstanceEndpoint.objects.bulk_create(endpoints)
        vm.private_ips = backend_vm.private_ips
        vm.public_ips = backend_vm.public_ips
        vm.save(update_fields=['private_ips', 'public_ips'])

    def get_vm(self, vm_id):
        try:
            vm = next(vm for vm in self.manager.list_nodes(self.cloud_service_name) if vm.id == vm_id)
            # XXX: libcloud seems doesn't map size properly
            vm.size = self.get_size(vm.extra['instance_size'])
            return vm
        except (StopIteration, LibcloudError) as e:
            six.reraise(AzureBackendError, e.message or "Virtual machine doesn't exist")

    def get_size(self, size_id):
        try:
            return next(s for s in self.manager.list_sizes() if s.id == size_id)
        except (StopIteration, LibcloudError) as e:
            six.reraise(AzureBackendError, e)

    def get_image(self, image_id):
        try:
            return next(s for s in self.manager.list_images() if s.id == image_id)
        except (StopIteration, LibcloudError) as e:
            six.reraise(AzureBackendError, e.message or "Image doesn't exist")

    def get_storage(self, storage_name):
        try:
            return next(s for s in self.manager.ex_list_storage_services() if s.service_name == storage_name)
        except (StopIteration, LibcloudError) as e:
            six.reraise(AzureBackendError, e.message or "Storage doesn't exist")

    def get_storage_name(self, cloud_service_name=None):
        if not cloud_service_name:
            cloud_service_name = self.cloud_service_name
        # Storage account name must be between 3 and 24 characters in length
        # and use numbers and lower-case letters only
        return re.sub(r'[\W_-]+', '', cloud_service_name.lower())[:24]

    def get_resources_for_import(self):
        if not self.cloud_service_name:
            raise AzureBackendError(
                "Resources could be fetched only for specific cloud service, "
                "please supply project_uuid query argument")

        cur_vms = models.VirtualMachine.objects.all().values_list('backend_id', flat=True)
        try:
            vms = self.manager.list_nodes(self.cloud_service_name)
        except LibcloudError as e:
            six.reraise(AzureBackendError, e)

        return [{
            'id': vm.id,
            'name': vm.name,
            'flavor_name': vm.extra.get('instance_size')
        } for vm in vms if vm.id not in cur_vms]

    def get_managed_resources(self):
        try:
            ids = []
            services = self.manager.ex_list_cloud_services()
            for service in services:
                for node in self.manager.list_nodes(service.service_name):
                    ids.append(node.id)
        except LibcloudError:
            return []
        return models.VirtualMachine.objects.filter(backend_id__in=ids)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-07 10:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('waldur_jira', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='resource_content_type',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='jira_issues', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='issue',
            name='resource_object_id',
            field=models.PositiveIntegerField(null=True, blank=True),
        ),
    ]



from django.apps import AppConfig


class CostPlanningConfig(AppConfig):
    name = 'waldur_cost_planning'
    verbose_name = 'Cost planning'

    def ready(self):
        from .plugins import digitalocean, openstack_tenant, aws, azure  # noqa: F401



import collections

Account = collections.namedtuple('Account', ['name', 'description', 'organization'])
Association = collections.namedtuple('Association', ['account', 'user', 'value'])


class Quotas(object):
    def __init__(self, cpu=0, gpu=0, ram=0, deposit=0):
        self.cpu = cpu
        self.gpu = gpu
        self.ram = ram
        self.deposit = deposit

    def __add__(self, other):
        return Quotas(
            self.cpu + other.cpu,
            self.gpu + other.gpu,
            self.ram + other.ram,
            self.deposit + other.deposit,
        )



from __future__ import unicode_literals

from django.contrib import admin

from . import models


class ServiceProviderAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'customer', 'created')


class AttributeOptionInline(admin.TabularInline):
    model = models.AttributeOption


class AttributeAdmin(admin.ModelAdmin):
    inlines = [AttributeOptionInline]


class ScreenshotsInline(admin.TabularInline):
    model = models.Screenshots


class PlansInline(admin.TabularInline):
    model = models.Plan


class OfferingAdmin(admin.ModelAdmin):
    inlines = [ScreenshotsInline, PlansInline]


admin.site.register(models.ServiceProvider, ServiceProviderAdmin)
admin.site.register(models.Category)
admin.site.register(models.Offering, OfferingAdmin)
admin.site.register(models.Section)
admin.site.register(models.Attribute, AttributeAdmin)
admin.site.register(models.Screenshots)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4

import django.utils.timezone
from django.db import migrations, models
import model_utils.fields
import waldur_core.core.fields


def init_backup_snapshots(apps, schema_editor):
    Backup = apps.get_model('openstack', 'Backup')
    Snapshot = apps.get_model('openstack', 'Snapshot')
    for backup in Backup.objects.all():
        if backup.metadata.get('system_snapshot_id'):
            Snapshot.objects.create(
                uuid=uuid4().hex,
                size=backup.metadata.get('system_snapshot_size', 0),
                backend_id=backup.metadata.get('system_snapshot_id'),
                tenant=backup.tenant,
                service_project_link=backup.instance.service_project_link,
                name='Backup %s snapshot' % backup.uuid.hex,
                state=3,  # OK state
            )
        if backup.metadata.get('data_snapshot_id'):
            Snapshot.objects.create(
                uuid=uuid4().hex,
                size=backup.metadata.get('data_snapshot_size', 0),
                backend_id=backup.metadata.get('data_snapshot_id'),
                tenant=backup.tenant,
                service_project_link=backup.instance.service_project_link,
                name='Backup %s snapshot' % backup.uuid.hex,
                state=3,  # OK state
            )


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0016_backup_state'),
    ]

    operations = [
        migrations.AddField(
            model_name='backup',
            name='snapshots',
            field=models.ManyToManyField(related_name='backups', to='openstack.Snapshot'),
        ),
        migrations.AlterField(
            model_name='backup',
            name='tenant',
            field=models.ForeignKey(related_name='backups', to='openstack.Tenant'),
        ),
        migrations.RunPython(init_backup_snapshots),
        migrations.CreateModel(
            name='BackupRestoration',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('runtime_state', models.CharField(max_length=150, verbose_name='runtime state', blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='backup',
            name='instance',
            field=models.ForeignKey(related_name='backups', on_delete=django.db.models.deletion.PROTECT, to='openstack.Instance'),
        ),
        migrations.AddField(
            model_name='backuprestoration',
            name='backup',
            field=models.ForeignKey(related_name='restorations', to='openstack.Backup'),
        ),
        migrations.AddField(
            model_name='backuprestoration',
            name='flavor',
            field=models.ForeignKey(related_name='+', to='openstack.Flavor'),
        ),
        migrations.AddField(
            model_name='backuprestoration',
            name='instance',
            field=models.OneToOneField(related_name='+', to='openstack.Instance'),
        ),
        migrations.RenameField(
            model_name='drbackuprestoration',
            old_name='dr_backup',
            new_name='backup',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0004_instance'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='action',
            field=models.CharField(max_length=50, blank=True),
        ),
        migrations.AddField(
            model_name='instance',
            name='action_details',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='action',
            field=models.CharField(max_length=50, blank=True),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='action_details',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='volume',
            name='action',
            field=models.CharField(max_length=50, blank=True),
        ),
        migrations.AddField(
            model_name='volume',
            name='action_details',
            field=models.TextField(blank=True),
        ),
    ]



from django.db import models as django_models

from waldur_core.core.managers import GenericKeyMixin


class ResourceSlaManager(GenericKeyMixin, django_models.Manager):
    pass


class ResourceItemManager(GenericKeyMixin, django_models.Manager):
    pass


class ResourceSlaStateTransitionManager(GenericKeyMixin, django_models.Manager):
    pass



from waldur_core.core import managers as core_managers

from . import models


class ApplicationSummaryQuerySet(core_managers.SummaryQuerySet):
    @property
    def model(self):
        return models.ApplicationModel



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-10-25 09:56
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0052_customer_subnets'),
        ('experts', '0009_description_text_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='expertrequest',
            name='customer',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, 
                                    to='structure.Customer'),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='project_name',
            field=models.CharField(blank=True, max_length=150),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='project_uuid',
            field=models.CharField(blank=True, max_length=32),
        ),
        migrations.AlterField(
            model_name='expertrequest',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+',
                                    to='structure.Project'),
        ),
        migrations.AddField(
            model_name='expertcontract',
            name='team_customer',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='structure.Customer'),
        ),
        migrations.AddField(
            model_name='expertcontract',
            name='team_name',
            field=models.CharField(blank=True, max_length=150),
        ),
        migrations.AddField(
            model_name='expertcontract',
            name='team_uuid',
            field=models.CharField(blank=True, max_length=32),
        ),
        migrations.AlterField(
            model_name='expertcontract',
            name='team',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='structure.Project'),
        ),
    ]



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class MarketplaceConfig(AppConfig):
    name = 'waldur_mastermind.marketplace'
    verbose_name = 'Marketplace'

    def ready(self):
        from . import handlers, models

        signals.post_save.connect(
            handlers.create_screenshot_thumbnail,
            sender=models.Screenshots,
            dispatch_uid='waldur_mastermind.marketplace.create_screenshot_thumbnail',
        )



import datetime
import saml2

from saml2.entity_category.edugain import COC

from waldur_core.core import WaldurExtension


class SAML2Extension(WaldurExtension):

    class Settings:
        # Read more: https://github.com/rohe/pysaml2-3/blob/master/doc/howto/config.rst
        # For an example configuration refer to the packaging/etc/waldur/saml2.conf.py.example
        # wiki: https://opennode.atlassian.net/wiki/display/WD/Identity+providers
        WALDUR_AUTH_SAML2 = {
            # used for assigning the registration method to the user
            'name': 'saml2',
            # full path to the xmlsec1 binary program
            'xmlsec_binary': '/usr/bin/xmlsec1',
            # required for assertion consumer, single logout services and entity ID
            'base_url': '',
            # directory with attribute mapping
            'attribute_map_dir': '',
            # set to True to output debugging information
            'debug': False,
            # IdPs metadata XML files stored locally
            'idp_metadata_local': [],
            # IdPs metadata XML files stored remotely
            'idp_metadata_remote': [],
            # logging
            # empty to disable logging SAML2-related stuff to file
            'log_file': '',
            'log_level': 'INFO',
            # Indicates if the entity will sign the logout requests
            'logout_requests_signed': 'true',
            # Indicates if the authentication requests sent should be signed by default
            'authn_requests_signed': 'true',
            # Identifies the Signature algorithm URL according to the XML Signature specification
            # SHA1 is used by default
            'signature_algorithm': None,
            # Identifies the Message Digest algorithm URL according to the XML Signature specification
            # SHA1 is used by default
            'digest_algorithm': None,
            # Identified NameID format to use. None means default, empty string ("") disables addition of entity
            'nameid_format': None,
            # PEM formatted certificate chain file
            'cert_file': '',
            # PEM formatted certificate key file
            'key_file': '',
            # SAML attributes that are required to identify a user
            'required_attributes': [],
            # SAML attributes that may be useful to have but not required
            'optional_attributes': [],
            # mapping between SAML attributes and User fields
            'saml_attribute_mapping': {},
            # organization responsible for the service
            # you can set multilanguage information here
            'organization': {},

            # links to the entity categories
            'categories': [COC],

            # attributes required by CoC
            # https://wiki.refeds.org/display/CODE/SAML+2+Profile+for+the+Data+Protection+Code+of+Conduct
            'privacy_statement_url': 'http://example.com/#/privacy-policy/',
            'display_name': 'Service provider display name',
            'description': 'Service provider description',

            # mdpi attributes
            'registration_policy': 'http://example.com/#/registration-policy/',
            'registration_authority': 'http://example.com/#/registration-authority/',
            'registration_instant': datetime.datetime(2017, 1, 1).isoformat(),

            'ENABLE_SINGLE_LOGOUT': False,
            'ALLOW_TO_SELECT_IDENTITY_PROVIDER': True,
            'IDENTITY_PROVIDER_URL': None,
            'IDENTITY_PROVIDER_LABEL': None,
            'DEFAULT_BINDING': saml2.BINDING_HTTP_POST,
        }

        # These shouldn't be configurable by user -- see SAML2 section for details
        SAML_CREATE_UNKNOWN_USER = True
        SAML_DJANGO_USER_MAIN_ATTRIBUTE = 'username'

        SAML_ATTRIBUTE_MAPPING = WALDUR_AUTH_SAML2['saml_attribute_mapping']

        SAML_CONFIG = {
            # full path to the xmlsec1 binary program
            'xmlsec_binary': WALDUR_AUTH_SAML2['xmlsec_binary'],

            # your entity id, usually your subdomain plus the url to the metadata view
            'entityid': WALDUR_AUTH_SAML2['base_url'] + '/api-auth/saml2/metadata/',

            'entity_category': WALDUR_AUTH_SAML2['categories'],

            # directory with attribute mapping
            'attribute_map_dir': WALDUR_AUTH_SAML2['attribute_map_dir'],

            # this block states what services we provide
            'service': {
                # we are just a lonely SP
                'sp': {
                    # Indicates if the entity will sign the logout requests
                    'logout_requests_signed': WALDUR_AUTH_SAML2['logout_requests_signed'],
                    # Indicates if the authentication requests sent should be signed by default
                    'authn_requests_signed': WALDUR_AUTH_SAML2['authn_requests_signed'],

                    'endpoints': {
                        # url and binding to the assertion consumer service view
                        # do not change the binding or service name
                        'assertion_consumer_service': [
                            (WALDUR_AUTH_SAML2['base_url'] + '/api-auth/saml2/login/complete/',
                             saml2.BINDING_HTTP_POST),
                        ],
                        # url and binding to the single logout service view
                        # do not change the binding or service name
                        'single_logout_service': [
                            (WALDUR_AUTH_SAML2['base_url'] + '/api-auth/saml2/logout/complete/',
                             saml2.BINDING_HTTP_REDIRECT),
                            (WALDUR_AUTH_SAML2['base_url'] + '/api-auth/saml2/logout/complete/',
                             saml2.BINDING_HTTP_POST),
                        ],
                    },

                    'extensions': {
                        'mdui': {
                            'UIInfo': {
                                'display_name': {
                                    'lang': 'en',
                                    'text': WALDUR_AUTH_SAML2['display_name'],
                                },
                                'description': {
                                    'lang': 'en',
                                    'text': WALDUR_AUTH_SAML2['description'],
                                },
                                'privacy_statement_url': {
                                    'lang': 'en',
                                    'text': WALDUR_AUTH_SAML2['privacy_statement_url'],
                                },
                            },
                        },
                        'mdrpi': {
                            'RegistrationInfo': {
                                'registration_policy': {
                                    'lang': 'en',
                                    'text': WALDUR_AUTH_SAML2['registration_policy'],
                                },
                                'registrationAuthority': WALDUR_AUTH_SAML2['registration_authority'],
                                'registrationInstant': WALDUR_AUTH_SAML2['registration_instant'],
                            },
                        }
                    },

                    # attributes that this project needs to identify a user
                    'required_attributes': WALDUR_AUTH_SAML2['required_attributes'],

                    # attributes that may be useful to have but not required
                    'optional_attributes': WALDUR_AUTH_SAML2['optional_attributes'],
                },
            },

            # Use database metadata loader
            # See also: https://github.com/rohe/pysaml2/issues/216
            'metadata': [
                {
                    'class': 'waldur_auth_saml2.utils.DatabaseMetadataLoader',
                    'metadata': [('waldur_auth_saml2.utils.DatabaseMetadataLoader',)],
                },
            ],

            'organization': WALDUR_AUTH_SAML2['organization'],

            # set to 1 to output debugging information
            'debug': int(WALDUR_AUTH_SAML2['debug']),

            # signing
            'key_file': WALDUR_AUTH_SAML2['key_file'],  # private part
            'cert_file': WALDUR_AUTH_SAML2['cert_file'],  # public part

            'only_use_keys_in_metadata': False,
            'allow_unknown_attributes': True,

            'accepted_time_diff': 120,
        }

    @staticmethod
    def update_settings(settings):
        settings['AUTHENTICATION_BACKENDS'] += ('djangosaml2.backends.Saml2Backend',)
        if settings['WALDUR_AUTH_SAML2']['log_file'] != '':
            level = settings['WALDUR_AUTH_SAML2']['log_level'].upper()
            settings['LOGGING']['handlers']['file-saml2'] = {
                'class': 'logging.handlers.WatchedFileHandler',
                'filename': settings['WALDUR_AUTH_SAML2']['log_file'],
                'formatter': 'simple',
                'level': level,
            }

            settings['LOGGING']['loggers']['djangosaml2'] = {
                'handlers': ['file-saml2'],
                'propagate': True,
                'level': level,
            }

            settings['LOGGING']['loggers']['saml2'] = {
                'handlers': ['file-saml2'],
                'propagate': True,
                'level': level,
            }

        # Use the same keypairs for both signing and encryption.
        # Otherwise pysaml2 doesn't decrypt encrypted assertion.
        # See also: https://github.com/knaperek/djangosaml2/issues/22
        settings['SAML_CONFIG']['encryption_keypairs'] = [{
            'key_file': settings['WALDUR_AUTH_SAML2']['key_file'],
            'cert_file': settings['WALDUR_AUTH_SAML2']['cert_file'],
        }]

        # Implement backward-compatible style for remote metadata specification.
        for remote in settings['WALDUR_AUTH_SAML2']['idp_metadata_remote']:
            settings['SAML_CONFIG']['metadata'].append({
                'class': 'saml2.mdstore.MetaDataExtern',
                'metadata': [(remote['url'], remote['cert'])]
            })

    @staticmethod
    def get_public_settings():
        return [
            'ENABLE_SINGLE_LOGOUT',
            'ALLOW_TO_SELECT_IDENTITY_PROVIDER',
            'IDENTITY_PROVIDER_URL',
            'IDENTITY_PROVIDER_LABEL',
        ]

    @staticmethod
    def django_app():
        return 'waldur_auth_saml2'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'waldur-auth-saml2-sync-providers': {
                'task': 'waldur_auth_saml2.sync_providers',
                'schedule': timedelta(minutes=60),
                'args': (),
            },
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4

from django.db import migrations, models


def populate_instance_volumes(apps, schema_editor):
    Instance = apps.get_model('openstack', 'Instance')
    Volume = apps.get_model('openstack', 'Volume')
    for instance in Instance.objects.all():
        if instance.system_volume_id:
            system_volume = Volume.objects.create(
                uuid=uuid4().hex,
                tenant=instance.tenant,
                service_project_link=instance.service_project_link,
                bootable=True,
                size=instance.system_volume_size,
                backend_id=instance.system_volume_id,
                name='{0}-system'.format(instance.name[:143]),
                state=3,
            )
            instance.volumes.add(system_volume)
        if instance.data_volume_id:
            data_volume = Volume.objects.create(
                uuid=uuid4().hex,
                tenant=instance.tenant,
                service_project_link=instance.service_project_link,
                bootable=False,
                size=instance.data_volume_size,
                backend_id=instance.data_volume_id,
                name='{0}-data'.format(instance.name[:145]),
                state=3,
            )
            instance.volumes.add(data_volume)


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0013_add_dr_backups_to_schedule'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='volumes',
            field=models.ManyToManyField(related_name='instances', to='openstack.Volume'),
        ),
        migrations.RunPython(populate_instance_volumes),
        migrations.RemoveField(
            model_name='drbackup',
            name='instance_volumes',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='data_volume_id',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='data_volume_size',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='system_volume_id',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='system_volume_size',
        ),
    ]



from waldur_core.core import models as core_models, tasks as core_tasks, utils as core_utils
from waldur_core.structure import models as structure_models, filters as structure_filters

from . import models


def remove_ssh_keys_from_service(sender, structure, user, role, **kwargs):
    """ Remove user ssh keys if he doesn't have access to service anymore. """
    if isinstance(structure, structure_models.Project):
        lost_services = models.DigitalOceanService.objects.filter(customer__projects=structure)
    elif isinstance(structure, structure_models.Customer):
        lost_services = models.DigitalOceanService.objects.filter(customer=structure)
    else:
        return
    lost_settings = [service.settings for service in lost_services]
    visible_services = structure_filters.filter_queryset_for_user(models.DigitalOceanService.objects.all(), user)
    visible_settings = [service.settings for service in visible_services]
    settings_list = set([settings for settings in lost_settings if settings not in visible_settings])

    ssh_keys = core_models.SshPublicKey.objects.filter(user=user)
    for settings in settings_list:
        serialized_settings = core_utils.serialize_instance(settings)
        for ssh_key in ssh_keys:
            core_tasks.IndependentBackendMethodTask().delay(
                serialized_settings, 'remove_ssh_key', ssh_key.name, ssh_key.fingerprint)


def remove_ssh_key_from_service_settings_on_deletion(sender, instance, **kwargs):
    ssh_key = instance
    user = ssh_key.user
    services = structure_filters.filter_queryset_for_user(models.DigitalOceanService.objects.all(), user)
    settings_list = structure_models.ServiceSettings.objects.filter(digitaloceanservice=services)
    for settings in settings_list:
        serialized_settings = core_utils.serialize_instance(settings)
        core_tasks.IndependentBackendMethodTask().delay(
            serialized_settings, 'remove_ssh_key', ssh_key.name, ssh_key.fingerprint)



from waldur_core.logging.loggers import EventLogger, event_logger

from . import models


class ExpertRequestEventLogger(EventLogger):
    expert_request = models.ExpertRequest

    class Meta:
        event_types = (
            'expert_request_created',
            'expert_request_activated',
            'expert_request_cancelled',
            'expert_request_completed',
        )
        event_groups = {
            'experts': event_types,
        }


class ExpertBidEventLogger(EventLogger):
    expert_bid = models.ExpertBid

    class Meta:
        event_types = ('expert_bid_created',)
        event_groups = {
            'experts': event_types,
        }


event_logger.register('waldur_expert_request', ExpertRequestEventLogger)
event_logger.register('waldur_expert_bid', ExpertBidEventLogger)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-08-06 12:56
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0035_unique_floating_ip'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='internalip',
            unique_together=set([]),
        ),
    ]



from django.db.models import Q

from waldur_core.quotas import fields as quota_fields
from waldur_core.structure import models as structure_models

from . import models


QUOTA_NAME = 'nc_expert_count'


def add_quota_field():
    structure_models.Project.add_quota_field(
        name=QUOTA_NAME,
        quota_field=quota_fields.QuotaField()
    )
    structure_models.Customer.add_quota_field(
        name=QUOTA_NAME,
        quota_field=quota_fields.QuotaField()
    )


def update_project_quota(project):
    if not project:
        return
    valid_states = (models.ExpertRequest.States.ACTIVE, models.ExpertRequest.States.PENDING)
    query = Q(project=project, state__in=valid_states)
    count = models.ExpertRequest.objects.filter(query).count()
    project.set_quota_usage(QUOTA_NAME, count)


def update_customer_quota(customer):
    if not customer:
        return
    valid_states = (models.ExpertRequest.States.ACTIVE, models.ExpertRequest.States.PENDING)
    query = Q(contract__team__customer=customer, state__in=valid_states)
    count = models.ExpertRequest.objects.filter(query).count()
    customer.set_quota_usage(QUOTA_NAME, count)


def get_request_customer(request):
    try:
        return request.contract.team.customer
    except AttributeError:
        return None


def get_contract_customer(contract):
    try:
        return contract.team.customer
    except AttributeError:
        return None


def get_experts_count(scope):
    return scope.quotas.get(name=QUOTA_NAME).usage



from django.core.management.base import BaseCommand

from waldur_core.structure import models as structure_models

from waldur_mastermind.experts import quotas


class Command(BaseCommand):
    help = "Recalculate experts quota usage for each project."

    def handle(self, *args, **options):
        for project in structure_models.Project.objects.all():
            quotas.update_project_quota(project)

        for customer in structure_models.Customer.objects.all():
            quotas.update_customer_quota(customer)



from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0052_customer_subnets'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExpertProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('customer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Customer')),
            ],
            options={
                'verbose_name': 'Expert providers',
            },
        ),
    ]



from django_filters import OrderingFilter

from waldur_core.structure import filters as structure_filters

from . import models


class ImageFilter(structure_filters.BaseServicePropertyFilter):

    o = OrderingFilter(fields=('distribution', 'type'))

    class Meta(object):
        model = models.Image
        fields = structure_filters.BaseServicePropertyFilter.Meta.fields + ('distribution', 'type')


class SizeFilter(structure_filters.BaseServicePropertyFilter):

    class Meta(object):
        model = models.Size
        fields = structure_filters.BaseServicePropertyFilter.Meta.fields + ('cores', 'ram', 'disk')


class RegionFilter(structure_filters.BaseServicePropertyFilter):
    class Meta(structure_filters.BaseServicePropertyFilter.Meta):
        model = models.Region


class DropletFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Droplet



from celery import chain
from django.core import checks
from django.db.migrations.topological_sort import stable_topological_sort
from django.db.models import Model

from waldur_core.core import WaldurExtension
from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.structure.tasks import ConnectSharedSettingsTask


class ServiceSettingsCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, settings, serialized_settings, **kwargs):
        creation_tasks = [core_tasks.StateTransitionTask().si(serialized_settings, state_transition='begin_creating')]
        # connect settings to all customers if they are shared
        if settings.shared:
            creation_tasks.append(ConnectSharedSettingsTask().si(serialized_settings))
        # sync settings if they have not only global properties
        backend = settings.get_backend()
        if not backend.has_global_properties():
            creation_tasks.append(core_tasks.IndependentBackendMethodTask().si(serialized_settings, 'sync'))
        return chain(*creation_tasks)


class ServiceSettingsPullExecutor(core_executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, settings, serialized_settings, **kwargs):
        return core_tasks.IndependentBackendMethodTask().si(
            serialized_settings, 'sync', state_transition='begin_updating')


class ServiceSettingsConnectSharedExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, settings, serialized_settings, **kwargs):
        return ConnectSharedSettingsTask().si(serialized_settings)


class BaseCleanupExecutor(core_executors.BaseExecutor):
    """
    In order to delete project and related resources, we need to take into account three constraints:

    1) dependencies between applications;
    2) dependencies between resources;
    3) backend API calls to delete resources.

    All these steps are optional, because some applications don't have all these constraints.

    1) If `related_executor` is specified, then current executor is applied before related executor.

    2) Project's resources are specified by the `pre_models` field.
       It is assumed that each model class can be filtered by project.

    3) The value of `executors` field is list of tuples (model class, executor class).
       Executors are applied after resources specified by pre_models field are deleted.
       When executor is applied, all resources in the project are deleted using this executor.

    Consider for example:

    class OpenStackCleanupExecutor(structure_executors.BaseCleanupExecutor):
        executors = (
            (models.SecurityGroup, SecurityGroupDeleteExecutor),
            (models.FloatingIP, FloatingIPDeleteExecutor),
            (models.SubNet, SubNetDeleteExecutor),
            (models.Network, NetworkDeleteExecutor),
            (models.Tenant, TenantDeleteExecutor),
        )

    class OpenStackTenantCleanupExecutor(structure_executors.BaseCleanupExecutor):
        related_executor = openstack_executors.OpenStackCleanupExecutor

        pre_models = (
            models.SnapshotSchedule,
            models.BackupSchedule,
        )

        executors = (
            (models.Snapshot, SnapshotDeleteExecutor),
            (models.Backup, BackupDeleteExecutor),
            (models.Instance, InstanceDeleteExecutor),
            (models.Volume, VolumeDeleteExecutor),
        )
    """

    pre_models = []
    executors = []
    related_executor = None

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        for model_cls in cls.pre_models:
            model_cls.objects.filter(project=instance).delete()

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        """
        Delete each resource using specific executor.
        Convert executors to task and combine all deletion task into single sequential task.
        """
        cleanup_tasks = [
            ProjectResourceCleanupTask().si(
                core_utils.serialize_class(executor_cls),
                core_utils.serialize_class(model_cls),
                serialized_instance,
            )
            for (model_cls, executor_cls) in cls.executors
        ]

        if not cleanup_tasks:
            return core_tasks.EmptyTask()

        return chain(cleanup_tasks)


class ProjectResourceCleanupTask(core_tasks.Task):

    @classmethod
    def get_description(cls, executor, model, project, *args, **kwargs):
        return 'Delete "%s" resources for project "%s" using executor %s.' % (model, project, executor)

    def run(self, serialized_executor, serialized_model, serialized_project, *args, **kwargs):
        executor = core_utils.deserialize_class(serialized_executor)
        model_cls = core_utils.deserialize_class(serialized_model)
        project = core_utils.deserialize_instance(serialized_project)

        for resource in model_cls.objects.filter(project=project):
            executor.execute(resource, async=False, force=True, **kwargs)


class ProjectCleanupExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        executors = cls.get_executors()

        # Combine all executors into single sequential task
        # to cleanup related resources in correct order
        cleanup_tasks = [
            core_tasks.ExecutorTask().si(
                core_utils.serialize_class(executor),
                serialized_instance
            )
            for executor in executors
        ]

        if not cleanup_tasks:
            return core_tasks.EmptyTask()

        return chain(cleanup_tasks)

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        # Delete project itself if all task have been completed successfully
        return core_tasks.DeletionTask().si(serialized_instance)

    @classmethod
    def get_executors(cls):
        # Get cleanup executors from extensions
        executors = [
            extension.get_cleanup_executor()
            for extension in WaldurExtension.get_extensions()
        ]

        # Filter empty items from list because cleanup executor is optional
        executors = [executor for executor in executors if executor]

        # Apply topological sort with respect to dependencies between executors
        dependencies = {}
        for executor in executors:
            dependencies[executor] = set()
            if executor.related_executor:
                dependencies.setdefault(executor.related_executor, set())
                dependencies[executor.related_executor].add(executor)
        return stable_topological_sort(executors, dependencies)


def is_valid_executor(item):
    return (
        isinstance(item, tuple) and
        issubclass(item[0], Model) and
        issubclass(item[1], core_executors.BaseExecutor)
    )


def check_cleanup_executors(app_configs, **kwargs):
    errors = []

    for extension in WaldurExtension.get_extensions():
        cleanup_executor = extension.get_cleanup_executor()
        if not cleanup_executor:
            continue

        for model in cleanup_executor.pre_models:
            if not issubclass(model, Model):
                errors.append(
                    checks.Error(
                        'Invalid resource model is detected in project cleanup executor.',
                        obj=cleanup_executor,
                        id='waldur.E001',
                    )
                )

        for item in cleanup_executor.executors:
            if not is_valid_executor(item):
                errors.append(
                    checks.Error(
                        'Invalid resource executor is detected in project cleanup executor.',
                        obj=cleanup_executor,
                        id='waldur.E001',
                    )
                )

    return errors



from django.core.validators import MaxValueValidator
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models


@python_2_unicode_compatible
class BaseSecurityGroupRule(models.Model):
    TCP = 'tcp'
    UDP = 'udp'
    ICMP = 'icmp'

    CHOICES = (
        (TCP, 'tcp'),
        (UDP, 'udp'),
        (ICMP, 'icmp'),
    )

    protocol = models.CharField(max_length=4, blank=True, choices=CHOICES)
    from_port = models.IntegerField(validators=[MaxValueValidator(65535)], null=True)
    to_port = models.IntegerField(validators=[MaxValueValidator(65535)], null=True)
    cidr = models.CharField(max_length=32, blank=True)

    backend_id = models.CharField(max_length=128, blank=True)

    class Meta(object):
        abstract = True

    def __str__(self):
        return '%s (%s): %s (%s -> %s)' % \
               (self.security_group, self.protocol, self.cidr, self.from_port, self.to_port)


@python_2_unicode_compatible
class Port(core_models.BackendModelMixin, models.Model):
    # TODO: Use dedicated field: https://github.com/django-macaddress/django-macaddress
    mac_address = models.CharField(max_length=32, blank=True)
    ip4_address = models.GenericIPAddressField(null=True, blank=True, protocol='IPv4')
    ip6_address = models.GenericIPAddressField(null=True, blank=True, protocol='IPv6')
    backend_id = models.CharField(max_length=255, blank=True)

    class Meta(object):
        abstract = True

    def __str__(self):
        return self.ip4_address or self.ip6_address or 'Not initialized'

    @classmethod
    def get_backend_fields(cls):
        return super(Port, cls).get_backend_fields() + ('ip4_address', 'ip6_address', 'mac_address')


class BaseImage(structure_models.ServiceProperty):
    min_disk = models.PositiveIntegerField(default=0, help_text=_('Minimum disk size in MiB'))
    min_ram = models.PositiveIntegerField(default=0, help_text=_('Minimum memory size in MiB'))

    class Meta(structure_models.ServiceProperty.Meta):
        abstract = True

    @classmethod
    def get_backend_fields(cls):
        return super(BaseImage, cls).get_backend_fields() + ('min_disk', 'min_ram')



from __future__ import unicode_literals

from django.db import transaction

from waldur_core.core import utils as core_utils

from .log import event_logger
from . import tasks, models


def log_issue_save(sender, instance, created=False, **kwargs):
    if created:
        return

    event_logger.waldur_issue.info(
        'Issue {issue_key} has been updated.',
        event_type='issue_update_succeeded',
        event_context={
            'issue': instance,
        })


def log_issue_delete(sender, instance, **kwargs):
    event_logger.waldur_issue.info(
        'Issue {issue_key} has been deleted.',
        event_type='issue_deletion_succeeded',
        event_context={
            'issue': instance,
        })


def log_offering_created(sender, instance, created=False, **kwargs):
    if created:
        event_logger.waldur_offering.info(
            'Offering {offering_name} has been created.',
            event_type='offering_created',
            event_context={
                'offering': instance,
            })


def log_offering_deleted(sender, instance, **kwargs):
    event_logger.waldur_offering.info(
        'Offering {offering_name} has been deleted.',
        event_type='offering_deleted',
        event_context={
            'offering': instance,
        })


def log_offering_state_changed(sender, instance, **kwargs):
    state = instance.state
    if state != instance.tracker.previous('state'):
        event_logger.waldur_offering.info(
            'Offering state has changed to {offering_state}',
            event_type='offering_state_changed',
            event_context={
                'offering': instance,
            }
        )


def send_comment_added_notification(sender, instance, created=False, **kwargs):
    comment = instance

    if not created or not comment.is_public:
        return

    serialized_comment = core_utils.serialize_instance(comment)
    transaction.on_commit(lambda:
                          tasks.send_comment_added_notification.delay(serialized_comment))


def send_issue_updated_notification(sender, instance, created=False, **kwargs):
    if created or set(instance.tracker.changed()) == {models.Issue.assignee.field.attname, 'modified'} or \
            set(instance.tracker.changed()) == {'modified'}:
        return

    # Skip notification if issue is not created on backend yet.
    if not instance.backend_id:
        return

    # Skip notification if issue just has been created on backend.
    if 'backend_id' in instance.tracker.changed():
        return

    serialized_issue = core_utils.serialize_instance(instance)
    transaction.on_commit(lambda:
                          tasks.send_issue_updated_notification.delay(serialized_issue))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-05-17 09:51
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0016_project_template_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='action',
            field=models.CharField(blank=True, max_length=50),
        ),
        migrations.AddField(
            model_name='project',
            name='action_details',
            field=waldur_core.core.fields.JSONField(default={}),
        ),
    ]



from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt
from django_openid_auth import views as auth_view

from waldur_core.core.views import validate_authentication_method

from . import views

validate_openid = validate_authentication_method('ESTONIAN_ID')
login_begin = validate_openid(csrf_exempt(auth_view.login_begin))
login_complete = validate_openid(auth_view.login_complete)


urlpatterns = [
    url(r'^api-auth/openid/login/$', login_begin, name='openid-login'),
    url(r'^api-auth/openid/complete/$', login_complete, name='openid-complete'),
    url(r'^api-auth/openid/logo.gif$', auth_view.logo, name='openid-logo'),
    url(r'^api-auth/openid/login_completed/', views.login_completed),
]



from __future__ import unicode_literals

from django.conf import settings
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers, models as structure_models
from waldur_openstack.openstack import (
    apps as openstack_apps, models as openstack_models, serializers as openstack_serializers)
from waldur_openstack.openstack_tenant import apps as openstack_tenant_apps

from . import models


class PackageComponentSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.PackageComponent
        fields = ('type', 'amount', 'price')


class PackageTemplateSerializer(core_serializers.AugmentedSerializerMixin,
                                serializers.HyperlinkedModelSerializer):
    price = serializers.DecimalField(max_digits=22, decimal_places=10)
    monthly_price = serializers.DecimalField(max_digits=16, decimal_places=2)
    components = PackageComponentSerializer(many=True)
    category = serializers.ReadOnlyField(source='get_category_display')

    class Meta(object):
        model = models.PackageTemplate
        fields = (
            'url', 'uuid', 'name', 'description', 'service_settings',
            'price', 'monthly_price', 'icon_url', 'components', 'category', 'archived',
            'product_code', 'article_code',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'service_settings': {'lookup_field': 'uuid'},
        }


def _check_template_service_settings(serializer, template):
    """ Template service settings should be in state OK and has type OpenStack """
    if template.service_settings.type != openstack_apps.OpenStackConfig.service_name:
        raise serializers.ValidationError(_('Template should be related to OpenStack service settings.'))
    elif template.service_settings.state != structure_models.ServiceSettings.States.OK:
        raise serializers.ValidationError(_('Template\'s settings must be in OK state.'))
    return template


def _get_template_quotas(template):
    components = {c.type: c.amount for c in template.components.all()}
    mapping = models.OpenStackPackage.get_quota_to_component_mapping()
    return {
        quota_name: components[component_type]
        for quota_name, component_type in mapping.items()
    }


def _apply_quotas(target, quotas):
    for name, limit in quotas.items():
        target.set_quota_limit(name, limit)


def _set_tenant_quotas(tenant, template):
    quotas = _get_template_quotas(template)
    _apply_quotas(tenant, quotas)


def _set_related_service_settings_quotas(tenant, template):
    quotas = _get_template_quotas(template)
    for target in structure_models.ServiceSettings.objects.filter(scope=tenant):
        _apply_quotas(target, quotas)


def _set_tenant_extra_configuration(tenant, template):
    tenant.extra_configuration = {
        'package_name': template.name,
        'package_uuid': template.uuid.hex,
        'package_category': template.get_category_display(),
    }
    for component in template.components.all():
        tenant.extra_configuration[component.type] = component.amount
    tenant.save()


def _has_access_to_package(user, spl):
    """ Staff and owner always have access to package. Manager - only if correspondent flag is enabled """
    manager_can_create = settings.WALDUR_OPENSTACK['MANAGER_CAN_MANAGE_TENANTS']
    return (
        user.is_staff or
        spl.service.customer.has_user(user, structure_models.CustomerRole.OWNER) or
        (manager_can_create and spl.project.has_user(user, structure_models.ProjectRole.MANAGER))
    )


class OpenStackPackageCreateSerializer(openstack_serializers.TenantSerializer):
    template = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='package-template-detail',
        write_only=True,
        queryset=models.PackageTemplate.objects.all())
    skip_connection_extnet = serializers.BooleanField(default=False)

    class Meta(openstack_serializers.TenantSerializer.Meta):
        fields = openstack_serializers.TenantSerializer.Meta.fields + ('template', 'skip_connection_extnet', )

    def _validate_service_project_link(self, spl):
        # It should be possible for owner to create package but impossible to create a package directly.
        # So we need to ignore tenant spl validation.

        user = self.context['request'].user
        if not _has_access_to_package(user, spl):
            raise serializers.ValidationError(_('You do not have permissions to create package for given project.'))
        return spl

    def validate_template(self, template):
        template = _check_template_service_settings(self, template)

        if template.archived:
            raise serializers.ValidationError(_('New package cannot be created for archived template.'))

        return template

    def validate(self, attrs):
        """ Additionally check that template and service project link belong to the same service settings """
        template = attrs['template']
        attrs = super(OpenStackPackageCreateSerializer, self).validate(attrs)
        spl = attrs['service_project_link']
        self._validate_service_project_link(spl)
        if spl.service.settings != template.service_settings:
            raise serializers.ValidationError(
                _('Template and service project link should be connected to the same service settings.'))
        return attrs

    @transaction.atomic
    def create(self, validated_data):
        """ Create tenant and service settings from it """
        template = validated_data.pop('template')
        tenant = super(OpenStackPackageCreateSerializer, self).create(validated_data)
        _set_tenant_quotas(tenant, template)
        _set_tenant_extra_configuration(tenant, template)

        # service settings are created on tenant creation
        service_settings = structure_models.ServiceSettings.objects.get(
            scope=tenant,
            type=openstack_tenant_apps.OpenStackTenantConfig.service_name,
        )
        package = models.OpenStackPackage.objects.create(
            tenant=tenant,
            template=template,
            service_settings=service_settings,
        )
        return package


class OpenStackPackageSerializer(core_serializers.AugmentedSerializerMixin,
                                 serializers.HyperlinkedModelSerializer):
    name = serializers.CharField(source='tenant.name', read_only=True)
    description = serializers.CharField(source='tenant.description', read_only=True)

    class Meta(object):
        model = models.OpenStackPackage
        fields = ('url', 'uuid', 'name', 'description', 'template', 'tenant', 'service_settings',)

        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'template': {'lookup_field': 'uuid', 'view_name': 'package-template-detail', 'read_only': True},
            'tenant': {'lookup_field': 'uuid', 'view_name': 'openstack-tenant-detail', 'read_only': True},
            'service_settings': {'lookup_field': 'uuid', 'read_only': True},
        }


class OpenStackPackageChangeSerializer(structure_serializers.PermissionFieldFilteringMixin, serializers.Serializer):
    package = serializers.HyperlinkedRelatedField(
        view_name='openstack-package-detail',
        lookup_field='uuid',
        queryset=models.OpenStackPackage.objects.all()
    )
    template = serializers.HyperlinkedRelatedField(
        view_name='package-template-detail',
        lookup_field='uuid',
        queryset=models.PackageTemplate.objects.all()
    )

    def get_filtered_field_names(self):
        return 'package',

    validate_template = _check_template_service_settings

    def validate_package(self, package):
        spl = package.tenant.service_project_link
        user = self.context['request'].user

        if package.tenant.state != openstack_models.Tenant.States.OK:
            raise serializers.ValidationError(_('Package\'s tenant must be in OK state.'))

        if not _has_access_to_package(user, spl):
            raise serializers.ValidationError(_('You do not have permissions to extend given package.'))

        return package

    def validate(self, attrs):
        package = attrs['package']
        new_template = attrs['template']
        if package.tenant.service_project_link.service.settings != new_template.service_settings:
            raise serializers.ValidationError(
                _('Template and package\'s tenant should be connected to the same service settings.'))

        if package.template == new_template:
            raise serializers.ValidationError(
                _('New package template cannot be the same as package\'s current template.'))

        usage = package.get_quota_usage()
        old_components = {component.type: component.amount for component in package.template.components.all()}
        for component in new_template.components.all():
            if component.type not in old_components:
                raise serializers.ValidationError(
                    _('Template\'s components must be the same as package template\'s components'))
            if component.type in usage and usage[component.type] > component.amount:
                msg = _("Current usage of {0} quota is greater than new template's {0} component.")
                raise serializers.ValidationError(msg.format(component.get_type_display()))
        return attrs

    @transaction.atomic
    def create(self, validated_data):
        package = validated_data['package']
        new_template = validated_data['template']
        service_settings = package.service_settings

        tenant = package.tenant
        _set_tenant_quotas(tenant, new_template)
        _set_related_service_settings_quotas(tenant, new_template)
        _set_tenant_extra_configuration(tenant, new_template)

        package.delete()
        new_package = models.OpenStackPackage.objects.create(
            template=new_template,
            service_settings=service_settings,
            tenant=tenant
        )

        return new_package


class OpenStackPackageAssignSerializer(serializers.Serializer):
    template = serializers.HyperlinkedRelatedField(
        view_name='package-template-detail',
        lookup_field='uuid',
        write_only=True,
        queryset=models.PackageTemplate.objects.all()
    )
    tenant = serializers.HyperlinkedRelatedField(
        view_name='openstack-tenant-detail',
        lookup_field='uuid',
        write_only=True,
        queryset=openstack_models.Tenant.objects.all(),
    )

    def validate_template(self, template):
        template = _check_template_service_settings(self, template)

        if template.archived:
            raise serializers.ValidationError(_('Package cannot be assigned for archived template.'))

        return template

    def validate_tenant(self, tenant):
        if models.OpenStackPackage.objects.filter(tenant=tenant).exists():
            raise serializers.ValidationError(
                _('Package for tenant already exists. '
                  'Please use change package operation instead.'))
        return tenant

    def validate(self, attrs):
        attrs = super(OpenStackPackageAssignSerializer, self).validate(attrs)
        spl = attrs['tenant'].service_project_link
        template = attrs['template']
        if spl.service.settings != template.service_settings:
            raise serializers.ValidationError(
                _('Template and service project link should be connected to the same service settings.'))

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        tenant = validated_data['tenant']
        template = validated_data['template']
        _set_tenant_quotas(tenant, template)
        _set_tenant_extra_configuration(tenant, template)
        service_settings = structure_models.ServiceSettings.objects.get(
            scope=tenant,
            type=openstack_tenant_apps.OpenStackTenantConfig.service_name,
        )
        return models.OpenStackPackage.objects.create(
            tenant=tenant,
            template=template,
            service_settings=service_settings,
        )



from __future__ import unicode_literals

import logging

from celery import shared_task
from django.conf import settings

from waldur_core.core import utils as core_utils
from waldur_core.core.utils import broadcast_mail
from waldur_core.structure import models as structure_models
from waldur_mastermind.support.tasks import _send_issue_notification

from . import models

logger = logging.getLogger(__name__)


def get_request_customer_link(request, customer):
    return settings.WALDUR_EXPERTS['REQUEST_CUSTOMER_LINK_TEMPLATE'].format(
        request_uuid=request.uuid.hex,
        customer_uuid=customer.uuid.hex
    )


def get_request_project_link(request):
    return settings.WALDUR_EXPERTS['REQUEST_PROJECT_LINK_TEMPLATE'].format(
        request_uuid=request.uuid.hex,
        project_uuid=request.project.uuid.hex
    )


@shared_task(name='waldur_mastermind.experts.send_new_request')
def send_new_request(request_uuid):
    """
    Send email notification about new expert request.
    """

    request = models.ExpertRequest.objects.get(uuid=request_uuid)

    enabled_providers = models.ExpertProvider.objects.filter(enable_notifications=True)
    customers = list(enabled_providers.values_list('customer', flat=True))
    customers = structure_models.Customer.objects.filter(pk__in=customers)
    for customer in customers:
        users = customer.get_owners()
        extra_context = {
            'request_link': get_request_customer_link(request, customer)
        }
        send_request_mail('new_request', request, users, extra_context)


@shared_task(name='waldur_mastermind.experts.send_new_bid')
def send_new_bid(bid_uuid):
    """
    Send email notification about new bid.
    """

    bid = models.ExpertBid.objects.get(uuid=bid_uuid)
    users = bid.request.customer.get_owners()
    extra_context = {
        'bid': bid,
        'request_link': get_request_project_link(bid.request)
    }
    send_request_mail('new_bid', bid.request, users, extra_context)


@shared_task(name='waldur_mastermind.experts.send_new_contract')
def send_new_contract(request_uuid):
    """
    Send email notification about accepted expert request.
    """

    request = models.ExpertRequest.objects.get(uuid=request_uuid)
    users = request.customer.get_owners()
    extra_context = {
        'request_link': get_request_project_link(request)
    }
    send_request_mail('contract', request, users, extra_context)


def send_request_mail(event_type, request, users, extra_context=None):
    """
    Shorthand to send email notification about expert request event.
    """

    recipient_list = list(users.exclude(email='').values_list('email', flat=True))

    context = dict(
        request=request,
        customer_name=request.project.customer.name,
        project_name=request.project.name,
        currency_name=settings.WALDUR_EXPERTS['CURRENCY_NAME'],
        site_name=settings.WALDUR_CORE['SITE_NAME'],
    )
    if extra_context:
        context.update(extra_context)

    logger.debug('About to send expert request {request_name} to {recipient_list}.'.format(
        request_name=request.name,
        recipient_list=', '.join(recipient_list)
    ))

    broadcast_mail('experts', event_type, context, recipient_list)


@shared_task
def create_pdf_contract(contract_id):
    contract = models.ExpertContract.objects.get(pk=contract_id)
    contract.create_file()


@shared_task(name='waldur_mastermind.experts.send_expert_comment_added_notification')
def send_expert_comment_added_notification(serialized_comment):
    # Send Expert notifications
    comment = core_utils.deserialize_instance(serialized_comment)
    expert_request = comment.issue.expertrequest_set.first()
    expert_contract = getattr(expert_request, 'contract', None)

    if expert_contract:
        experts = expert_contract.team.get_users(structure_models.ProjectRole.ADMINISTRATOR)

        if comment.author.user not in experts:
            for expert in experts:
                _send_issue_notification(comment.issue, 'comment_added', receiver=expert)
